(*Author: Giuliano Losa *)

header {* Proof that the composition of two instances of the ALM automaton behaves like a single instance of the ALM automaton*}

theory CompositionCorrectness
imports ALM
begin

declare 
  starts_of_def[simp] 
  hide_def[simp] 
  composeALMs_def[simp] 
  ALM_ioa_def[simp] 
  par_def[simp] 
  ALM_start_def[simp]
  trans_of_def[simp] 
  actions_def[simp] 
  asig_inputs_def[simp] 
  asig_outputs_def[simp] 
  asig_internals_def[simp] 
  asig_of_def[simp] 
  ALM_asig_def[simp]
  ALM_trans_def[simp]
  split_if_asm [split]

subsection {*A case split useful in the proofs*}

definition either1 :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow>  
  (nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool) \<Rightarrow> 
  ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "either1 i id1 id2 tr s t s' t' \<equiv> 
    (i < id1 \<and> tr 0 id1 s s'\<and> t = t') \<or> (i \<ge> id1 \<and> i < id2 \<and> tr id1 id2 t t' \<and> s = s')"
declare either1_def[simp]

definition either2 :: "nat \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow>  
  (nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool) \<Rightarrow> 
  ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "either2 i id1 id2 tr s t s' t' \<equiv> 
    (i = 0 \<and> tr 0 id1 s s' \<and> t = t') \<or> (i = id1 \<and> tr id1 id2 t t' \<and> s = s')"
declare either2_def[simp]

definition both :: "nat \<Rightarrow> nat \<Rightarrow> 
  (nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool) \<Rightarrow> 
  ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "both id1 id2 tr s t s' t' \<equiv> (tr 0 id1 s s') \<and> (tr id1 id2 t t')"
declare both_def[simp]

fun trans_cases_act_p :: "nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_action \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "trans_cases_act_p id1 id2 s t (Invoke c r) s' t' = (both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s t s' t')" |
  "trans_cases_act_p id1 id2 s t (Switch c i h r) s' t' = 
    ((i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans2 id1 id2 c id1 h r s s' \<and> Switch_trans1 id1 id2 c id1 h r t t')
    \<or> (i = id2 \<and> s = s' \<and> Switch_trans2 id1 id2 c id2 h r t t'))" |
  "trans_cases_act_p id1 id2 s t (Commit c i h) s' t' =
    either1 i id1 id2 (\<lambda> m n . Commit_trans m n c i h) s t s' t'" |
  "trans_cases_act_p id1 id2 s t (Linearize i h) s' t' = 
    either2 i id1 id2 (\<lambda> m n . Linearize_trans m n i h) s t s' t'" |
  "trans_cases_act_p id1 id2 s t (Initialize i h) s' t' =  
    (either2 i id1 id2 (\<lambda> m n . Initialize_trans m n i h) s t s' t')" |
  "trans_cases_act_p id1 id2 s t (Abort i) s' t' =  
    (either2 i id1 id2 (\<lambda> m n . Abort_trans m n i) s t s' t')"

lemma  trans_cases_act:
  -- {*A rule for decomposing proofs*}
  fixes id1 id2 s t a s' t'
  assumes "id1 \<noteq> 0" and "id1 < id2" and in_trans_comp:"(s,t) -(a::ALM_action)--composeALMs id1 id2-> (s',t')"
  shows "trans_cases_act_p id1 id2 s t a s' t'"
  using assms by (force simp: Switch_trans1_def Switch_trans2_def)
  
fun trans_cases_p :: "nat \<Rightarrow> nat \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> ALM_state \<Rightarrow> bool"
where
  "trans_cases_p id1 id2 s t s' t' =
    (\<exists> c i h r .
      both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s t s' t'
      \<or> (i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans2 id1 id2 c id1 h r s s' \<and> Switch_trans1 id1 id2 c id1 h r t t')
      \<or> (i = id2 \<and> s = s' \<and> Switch_trans2 id1 id2 c id2 h r t t')
      \<or> (i < id1 \<and> Commit_trans 0 id1 c i h s s'\<and> t = t')
      \<or> (i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h t t' \<and> s = s')
      \<or> (i = 0 \<and> Linearize_trans 0 id1 0 h s s' \<and> t = t') 
      \<or> (i = id1 \<and> Linearize_trans id1 id2 id1 h t t' \<and> s = s')
      \<or> (i = 0 \<and> Initialize_trans 0 id1 0 h s s' \<and> t = t') 
      \<or> (i = id1 \<and> Initialize_trans id1 id2 id1 h t t' \<and> s = s')
      \<or> (i = 0 \<and> Abort_trans 0 id1 0 s s' \<and> t = t') 
      \<or> (i = id1 \<and> Abort_trans id1 id2 id1 t t' \<and> s = s'))"
  
lemma trans_cases:
  fixes id1 id2 s t a s' t' P
  assumes "id1 \<noteq> 0" and "id1 < id2" and "(s,t) -a--composeALMs id1 id2-> (s',t')"
  shows "trans_cases_p id1 id2 s t s' t'"
proof -
  from assms and trans_cases_act have "trans_cases_act_p id1 id2 s t a s' t'" by blast
  thus ?thesis by (cases a, auto)
qed

subsection {*Invariants of a single ALM instance*}

fun P1a :: "(ALM_state * ALM_state) \<Rightarrow> bool" 
  where
  -- {*In ALM 1, a pending request of client c has client c as sender*}
  "P1a (s1,s2) = (\<forall> c . phase s1 c \<in> {Pending, Aborted} --> request_snd (pending s1 c) = c)"

fun P1b :: "(ALM_state * ALM_state) \<Rightarrow> bool"
  where
  -- {*In ALM 2, a pending request of client c has client c as sender*}
  "P1b (s1,s2) = (\<forall> c . phase s2 c \<noteq> Sleep --> request_snd (pending s2 c) = c)"

fun P2 :: "(ALM_state * ALM_state) \<Rightarrow> bool" where
  "P2 (s1,s2) = ((\<forall> c . phase s2 c = Sleep) \<longrightarrow> (\<not> initialized s2 \<and> hist s2 = []))"

fun P3 :: "(ALM_state * ALM_state) \<Rightarrow> bool" where
  "P3 (s1,s2) = (\<forall> c . (phase s2 c = Ready \<longrightarrow> initialized s2))"

fun P4 :: "(ALM_state * ALM_state) => bool" 
  where
  -- {*The set of init histories of ALM 2 is empty when no client ever invoked anything*}
  "P4 (s1,s2) = ((\<forall> c . phase s2 c = Sleep) = (initHists s2 = {}))"

fun P5 :: "(ALM_state * ALM_state) \<Rightarrow> bool" 
  -- {*In ALM 1 a client never sleeps*}
  where
  "P5 (s1,s2) = (\<forall> c . phase s1 c \<noteq> Sleep)"

subsection {*Invariants of the composition of two ALM instances*}

fun P6 :: "(ALM_state * ALM_state) => bool"
  -- {*Non-interference accross instances*}
  where
  "P6 (s1,s2) = ((~ aborted s1 \<longrightarrow> (ALL c . phase s2 c = Sleep)) 
    \<and> (ALL c . phase s1 c \<noteq> Aborted = (phase s2 c = Sleep)))"

fun P7 :: "(ALM_state * ALM_state) => bool" 
  -- {*Before initialization of the ALM 2, pending requests are the same as in ALM 1 and no new requests may be accepted (phase is not Ready)*}
  where
  "P7 (s1,s2) = (\<forall> c . phase s1 c = Aborted \<and> \<not> initialized s2 
    \<longrightarrow> (pending s2 c = pending s1 c \<and> phase s2 c \<in> {Pending, Aborted}))"

fun P8 :: "(ALM_state * ALM_state) => bool" 
  -- {*Init histories of ALM 2 are built from the history of ALM 1 plus pending requests of ALM 1*}
  where
  "P8 (s1,s2) = (\<forall> h \<in> initHists s2 . h \<in> postfix_all (hist s1) (linearizations (pendingReqs s1)))"

fun P9 :: "(ALM_state * ALM_state) \<Rightarrow> bool" 
  -- {*ALM 2 does not abort before ALM 1 aborts*}
  where
  "P9 (s1,s2)= (aborted s2 \<longrightarrow> aborted s1)"

fun P10 :: "(ALM_state * ALM_state) => bool" 
  -- {*ALM 1 is always initialized and when ALM 2 is not initialized its history is empty*}
  where
  "P10 (s1,s2) = (initialized s1 \<and> (\<not> initialized s2 \<longrightarrow> (hist s2 = [])))"

fun P11 :: "(ALM_state * ALM_state) => bool" 
  where
  -- {*After ALM 2 has been invoked and before it is initialized, any request found in init histories after their longest common prefix is pending in ALM 1*}
  "P11 (s1,s2) = (((\<exists> c . phase s2 c \<noteq> Sleep) \<and> \<not> initialized s2) 
    \<longrightarrow> initValidReqs s2 \<subseteq> pendingReqs s1)"

fun P12:: "(ALM_state * ALM_state) => bool" 
  where
  -- {*After ALM 2 has been invoked and before it is initialized, the longest common prefix of the init histories of ALM 2 is buit from appending a set of request pending in ALM 1 to the history of ALM 1*}
  "P12 (s1,s2) = ((\<exists> c . phase s2 c \<noteq> Sleep) 
    \<longrightarrow> (\<exists> rs . l_c_p (initHists s2) = rs @ (hist s1) \<and> set rs \<subseteq> pendingReqs s1 \<and> distinct rs))"

fun P13 :: "(ALM_state * ALM_state) => bool" 
  where
  -- {*After ALM 2 has been invoked and before it is initialized, any history that may be chosen at initialization is a valid linearization of the concurrent history of ALM 1*}
  "P13 (s1,s2) = (((\<exists> c . phase s2 c \<noteq> Sleep) \<and> \<not> initialized s2) 
    \<longrightarrow> postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2)) \<subseteq> postfix_all (hist s1) (linearizations (pendingReqs s1)))"

fun P14 :: "(ALM_state * ALM_state) => bool" 
  where
  -- {*The history of ALM 1 is a postfix of the history of ALM 2 and requests appearing in ALM 2 after the history of ALM 1 are not in the history of ALM 1*}
  "P14 (s1,s2) = ((hist s2 \<noteq> [] \<or> initialized s2) \<longrightarrow> (\<exists> rs . 
          hist s2 = rs @ (hist s1) 
          \<and> set rs \<inter> set (hist s1) = {}))"

fun P15 :: "(ALM_state * ALM_state) => bool" 
  where
  -- {*A client that hasn't yet invoked ALM 2 has no request commited in ALM 2 except for its pending request*}
  "P15 (s1,s2) = (\<forall> r . let c = request_snd r in phase s2 c = Sleep \<and> r \<in> set (hist s2) 
    \<longrightarrow> (r \<in> set (hist s1) \<or> r \<in> pendingReqs s1))"
    
subsection {*Proofs of invariance*}

lemma invariant_imp: "\<lbrakk>invariant ioa P; \<forall> s . P s \<longrightarrow> Q s\<rbrakk> \<Longrightarrow> invariant ioa Q"
by (simp add:invariant_def)

lemma dropWhile_lemma: "\<forall> ys . xs = ys @ zs \<and> hd zs = x \<and> zs \<noteq> [] \<and> x \<notin> set ys \<longrightarrow> dropWhile (\<lambda> x' . x' \<noteq> x) xs = zs"
  -- {*A useful lemma about truncating histories*}
proof  (induct xs, force)
  fix a xs
  assume "\<forall> ys . xs = ys @ zs \<and> hd zs = x \<and> zs \<noteq> [] \<and> x \<notin> set ys \<longrightarrow> dropWhile (\<lambda>x'. x' \<noteq> x) xs = zs"
  show "\<forall> ys . a # xs = ys @ zs \<and> hd zs = x \<and> zs \<noteq> [] \<and> x \<notin> set ys \<longrightarrow> dropWhile (\<lambda>x'. x' \<noteq> x) (a # xs) = zs"
  proof (rule allI, rule impI, cases "a = x")
    fix ys
    assume "a # xs = ys @ zs \<and> hd zs = x \<and> zs \<noteq> [] \<and> x \<notin> set ys" and "a = x"
    hence "x # xs = ys @ zs" and "x \<notin> set ys" and "hd zs = x" and "zs \<noteq> []" by auto
    from `x # xs = ys @ zs` and `x \<notin> set ys` have "ys = []"   by (metis hd.simps hd_append hd_in_set)
    with `a = x` and `x # xs = ys @ zs` show "dropWhile (\<lambda>x'. x' \<noteq> x) (a # xs) = zs" by auto
  next
    fix ys
    assume "a # xs = ys @ zs \<and> hd zs = x \<and> zs \<noteq> [] \<and> x \<notin> set ys" and "a \<noteq> x"
    hence "a # xs = ys @ zs" and "hd zs = x" and "zs \<noteq> []" and "x \<notin> set ys" by auto
    obtain ys' where "xs = ys' @ zs" and "x \<notin> set ys'"
    proof -
      from `a # xs = ys @ zs` and `hd zs = x` and `a \<noteq> x` obtain ys' where "ys = a # ys'" apply clarify by (metis Cons_eq_append_conv hd.simps)
      moreover with `x \<notin> set ys` have "x \<notin> set ys'" by auto
      moreover from `ys = a # ys'` and `a # xs = ys @ zs` have "xs = ys' @ zs" by auto
      ultimately show "(\<And>ys'. \<lbrakk>xs = ys' @ zs; x \<notin> set ys'\<rbrakk> \<Longrightarrow> thesis) \<Longrightarrow> thesis" by auto
    qed
    with `\<forall> ys . xs = ys @ zs \<and> hd zs = x \<and> zs \<noteq> [] \<and> x \<notin> set ys \<longrightarrow> dropWhile (\<lambda>x'. x' \<noteq> x) xs = zs` and `hd zs = x` and `zs \<noteq> []` have "dropWhile (\<lambda>x'. x' \<noteq> x) xs = zs" by auto
    with `a \<noteq> x` show "dropWhile (\<lambda>x'. x' \<noteq> x) (a # xs) = zs" by auto
  qed
qed

declare
  Invoke_trans_def[simp]
  Switch_trans1_def[simp]
  Switch_trans2_def[simp]
  Commit_trans_def[simp]
  Abort_trans_def[simp]
  Linearize_trans_def[simp]
  Initialize_trans_def[simp]
 
declare
  phase.split [split]
  phase.split_asm [split]
  request.split [split]
  request.split_asm [split]
  ALM_action.split [split]
  ALM_action.split_asm [split]

lemma P2_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P2"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P2 (s1, s2)" by simp 
}
moreover
{ fix s1 s2 s1' s2' act
  assume hyp: "P2 (s1, s2)" and in_trans_comp:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  hence "P2 (s1', s2')" using assms by auto
} 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P5_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P5"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P5 (s1, s2)" by simp 
} moreover
{ fix s1 s2 s1' s2' act
  assume "P5 (s1, s2)"
  and in_trans_comp:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  hence "P5 (s1', s2')" using assms by auto } 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P6_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P6"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P6 (s1, s2)" by simp 
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp: "P6 (s1, s2)" and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from trans and assms and trans_cases have cases:"trans_cases_p id1 id2 s1 s2 s1' s2'" by blast
  { fix c r
    assume "both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s1 s2 s1' s2'"
    with hyp have "P6 (s1', s2')"  by (simp, metis phase.distinct(11) phase.distinct(3), metis phase.distinct(3), metis phase.distinct(11))
  } moreover
  { fix  c i h r
    assume "i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans2 id1 id2 c id1 h r s1 s1' \<and> Switch_trans1 id1 id2 c id1 h r s2 s2'"
    with hyp have "P6 (s1', s2')" by simp
  } moreover
  { fix  c i h r
    assume "i = id2 \<and> s1 = s1' \<and> Switch_trans2 id1 id2 c id2 h r s2 s2'"
    with hyp have "P6 (s1', s2')" by fastforce
  } moreover
  { fix c i h
    assume "i < id1 \<and> Commit_trans 0 id1 c i h s1 s1'\<and> s2 = s2'"
    with hyp  have "P6 (s1', s2')" by force
  } moreover
  { fix c i h
    assume "i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1'"
    with hyp  have "P6 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = (0::nat) \<and> Linearize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp have "P6 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = id1 \<and> Linearize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P6 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = (0::nat) \<and> Initialize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp have "P6 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = id1 \<and> Initialize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P6 (s1', s2')" by fastforce
  } moreover
  { fix i
    assume "i = (0::nat) \<and> Abort_trans 0 id1 0 s1 s1' \<and> s2 = s2'"
    with hyp have "P6 (s1', s2')" by force
  } moreover
  { fix i
    assume "i = id1 \<and> Abort_trans id1 id2 id1 s2 s2' \<and> s1 = s1'"
    with hyp have "P6 (s1', s2')" by force
  }
  ultimately have "P6 (s1', s2')" using cases by (simp only:trans_cases_p.simps, metis)
}
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P9_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P9"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P9 (s1, s2)" by simp
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp:"P9 (s1, s2)" and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from reach and P6_invariant have P6:"P6 (s1, s2)" using assms by (metis (full_types) invariantE)
  from P6 and hyp and trans have "P9 (s1', s2')" using assms by auto } 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P10_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P10"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P10 (s1, s2)" by simp 
} moreover
{ fix s1 s2 s1' s2' act
  assume "P10 (s1, s2)"
  and in_trans_comp:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  hence "P10 (s1', s2')" using assms by auto } 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P3_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P3"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P3 (s1, s2)" by simp
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp:"P3 (s1, s2)" and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from reach and P10_invariant have P10:"P10 (s1, s2)" using assms by (metis (full_types) invariantE)
  from P10 and hyp and trans have "P3 (s1', s2')" using assms by auto } 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P7_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P7"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P7 (s1, s2)" by simp
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp:"P7 (s1, s2)" and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from reach and P6_invariant and P10_invariant have P6:"P6 (s1, s2)" and P10:"P10 (s1,s2)" 
    using assms by (metis (full_types) invariantE, metis (full_types) invariantE)
  from trans and assms and trans_cases have cases:"trans_cases_p id1 id2 s1 s2 s1' s2'" by blast
  { fix c r
    assume "both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s1 s2 s1' s2'"
    with hyp have "P7 (s1', s2')"  by force
  } moreover
  { fix  c i h r
    assume "i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans2 id1 id2 c id1 h r s1 s1' \<and> Switch_trans1 id1 id2 c id1 h r s2 s2'"
    with hyp and P6 have "P7 (s1', s2')" by auto
  } moreover
  { fix  c i h r
    assume "i = id2 \<and> s1 = s1' \<and> Switch_trans2 id1 id2 c id2 h r s2 s2'"
    with hyp have "P7 (s1', s2')" by fastforce
  } moreover
  { fix c i h
    assume "i < id1 \<and> Commit_trans 0 id1 c i h s1 s1'\<and> s2 = s2'"
    with hyp  have "P7 (s1', s2')" by force
  } moreover
  { fix c i h
    assume "i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1'"
    with hyp and P10 have "P7 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = (0::nat) \<and> Linearize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp have "P7 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = id1 \<and> Linearize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P7 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = (0::nat) \<and> Initialize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp have "P7 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = id1 \<and> Initialize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P7 (s1', s2')" by fastforce
  } moreover
  { fix i
    assume "i = (0::nat) \<and> Abort_trans 0 id1 0 s1 s1' \<and> s2 = s2'"
    with hyp have "P7 (s1', s2')" by force
  } moreover
  { fix i
    assume "i = id1 \<and> Abort_trans id1 id2 id1 s2 s2' \<and> s1 = s1'"
    with hyp have "P7 (s1', s2')" by force
  }
  ultimately have "P7 (s1', s2')" using cases by (simp only:trans_cases_p.simps, metis)
}
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P4_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P4"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P4 (s1, s2)" by simp 
} moreover
{ fix s1 s2 s1' s2' act
  assume "P4 (s1, s2)"
  and in_trans_comp:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  hence "P4 (s1', s2')" using assms by auto }
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed
(*
declare
  pendingReqs_def[simp]
  initValidReqs_def[simp]
*)
(* TODO *)
lemma P8_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P8"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P8 (s1, s2)" by simp
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp:"P8 (s1, s2)" and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from reach and P6_invariant and P10_invariant and P5_invariant and P4_invariant have P6:"P6 (s1, s2)" and P10:"P10 (s1,s2)" and P5:"P5 (s1, s2)" and P4:"P4 (s1,s2)" 
    using assms by (metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE)
  from trans and assms and trans_cases have cases:"trans_cases_p id1 id2 s1 s2 s1' s2'" by blast
  { fix c r
    assume invoke:"both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s1 s2 s1' s2'"
    { assume "s1' = s1"
      with invoke and hyp have "P8 (s1',s2')" by force
    } moreover
    { assume a:"s1' \<noteq> s1"
      from invoke and a have "pendingReqs s1 \<subseteq> pendingReqs s1'" by (force simp add:pendingReqs_def)
      moreover from invoke have "hist s1' = hist s1" by auto
      moreover from invoke have "initHists s2' = initHists s2" by force
      moreover note hyp
      ultimately have "P8 (s1', s2')" by (fastforce simp: linearizations_def postfix_all_def)
    }
    ultimately have "P8 (s1', s2')" by fast
  } moreover
  { fix  c i h r
    assume ai:"i = id1 \<and> id1 \<noteq> 0" and switch1:"Switch_trans2 id1 id2 c id1 h r s1 s1'" and switch2:"Switch_trans1 id1 id2 c id1 h r s2 s2'"
    { fix h1
      assume "h1 \<in> initHists s2'"
      have "h1 \<in> postfix_all (hist s1') (linearizations (pendingReqs s1'))"
      proof -
        { assume a:"h1 \<in> initHists s2"
          from switch1 have f1:"hist s1' = hist s1" and f2:"pendingReqs s1' = pendingReqs s1" by (auto simp: pendingReqs_def)
          from a and f1 and f2 and hyp have ?thesis by (force simp: postfix_all_def linearizations_def)
        } moreover
        { assume "h1 \<notin> initHists s2"
          with `h1 \<in> initHists s2'` and switch2 have "h1 = h" by simp
          with switch1 and P10 have "h1 \<in> postfix_all (hist s1) (linearizations (pendingReqs s1))" by force
          moreover from switch1 have "hist s1' = hist s1" and "pendingReqs s1' = pendingReqs s1" by (auto simp: pendingReqs_def)
          ultimately have ?thesis by auto
        }
        ultimately show ?thesis by fast
      qed
    }
    hence "P8 (s1', s2')" by simp
  } moreover
  { fix  c i h r
    assume "i = id2 \<and> s1 = s1' \<and> Switch_trans2 id1 id2 c id2 h r s2 s2'"
    with hyp have "P8 (s1', s2')" by fastforce
  } moreover
  { fix c i h
    assume "i < id1" and "Commit_trans 0 id1 c i h s1 s1'" and "s2 = s2'"
    hence "pendingReqs s1' = pendingReqs s1" and "hist s1' = hist s1" by (auto simp: pendingReqs_def)
    with hyp and `s2 = s2'` have "P8 (s1', s2')" by simp
  } moreover
  { fix c i h
    assume "i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1'"
    with hyp and P10 have "P8 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = (0::nat) \<and> Linearize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp and P6 and P4 have "P8 (s1', s2')" by force
  } moreover
  { fix i h 
    assume "i = id1 \<and> Linearize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P8 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = (0::nat) \<and> Initialize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp and P10 have "P8 (s1', s2')" by fastforce
  } moreover
  { fix i h 
    assume "i = id1 \<and> Initialize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P8 (s1', s2')" by fastforce
  } moreover
  { fix i
    assume "i = (0::nat) \<and> Abort_trans 0 id1 0 s1 s1' \<and> s2 = s2'"
    with hyp have "P8 (s1', s2')" by (force simp add:pendingReqs_def)
  } moreover
  { fix i
    assume "i = id1 \<and> Abort_trans id1 id2 id1 s2 s2' \<and> s1 = s1'"
    with hyp have "P8 (s1', s2')" by force
  }
  ultimately have "P8 (s1', s2')" using cases by (simp only:trans_cases_p.simps, metis)
}
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P12_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P12"
proof -
  { fix s1 s2
    assume P8:"P8 (s1, s2)" and P4:"P4 (s1, s2)"
    have "P12 (s1, s2)"
    proof -
      from P8 and P4 have f1:"\<forall> h \<in> initHists s2 . (\<exists> h' . h = h' @ (hist s1) \<and> set h' \<subseteq> pendingReqs s1 \<and> distinct h')" 
        by (auto simp add:postfix_all_def linearizations_def)
      { assume a:"\<exists> c . phase s2 c \<noteq> Sleep"
        have "\<exists> rs . l_c_p (initHists s2) = rs @ hist s1 \<and> set rs \<subseteq> pendingReqs s1 \<and> distinct rs"
        proof -
          from a and P4 have "initHists s2 \<noteq> {}" by force
          with l_c_p_lemma[of "initHists s2" "hist s1"] and f1 
            obtain rs where "l_c_p (initHists s2) = rs @ hist s1" by (metis suffixeq_def)
          moreover have "set rs \<subseteq> pendingReqs s1"
          proof -
            from `initHists s2 \<noteq> {}` obtain h where "h \<in> initHists s2" by auto
            with f1 obtain h' where "h = h' @ (hist s1) \<and> set h' \<subseteq> pendingReqs s1" by auto
            moreover from l_c_p_common_postfix[of "initHists s2"] and `h \<in> initHists s2` 
              obtain h'' where "h = h'' @ (l_c_p (initHists s2))" by (auto simp add:common_postfix_p_def suffixeq_def)
            moreover note `l_c_p (initHists s2) = rs @ hist s1`
            ultimately show ?thesis by auto
          qed
          moreover have "distinct rs"
          proof -
            from `initHists s2 \<noteq> {}` obtain h where "h \<in> initHists s2" by auto
            with f1  obtain h' where "h = h' @ (hist s1)" and "distinct h'" by auto
            with l_c_p_common_postfix[of "initHists s2"] and `h \<in> initHists s2` and `l_c_p (initHists s2) = rs @ hist s1` 
              obtain h'' where "h' = h'' @ rs"  by (fastforce simp: common_postfix_p_def suffixeq_def)
            with `distinct h'` show ?thesis by auto
          qed
          ultimately show ?thesis by blast
        qed
      }
      thus ?thesis by simp
    qed
  }
  thus ?thesis by (metis Automata.invariant_def P4_invariant P8_invariant PairE assms)
qed

lemma P11_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P11"
proof -
  { fix s1 s2
    assume P8:"P8 (s1, s2)" and P12:"P12 (s1, s2)" and P6:"P6 (s1, s2)" and P7:"P7 (s1, s2)"
    have "P11 (s1, s2)"
    proof -
      { fix x c h
        assume "phase s2 c \<noteq> Sleep"
        with P12 and P8 
          have initHists_prop:"\<forall> h \<in> initHists s2 . (\<exists> h' . h = h' @ (hist s1) \<and> set h' \<subseteq> pendingReqs s1)" 
          and lcp_prop:"\<exists> rs . l_c_p (initHists s2) = rs @ (hist s1)" 
          by (auto simp add: postfix_all_def linearizations_def)
        assume "x \<notin> set (l_c_p (initHists s2))" and "h \<in> initHists s2" and "x \<in> set h"
        from initHists_prop and `h \<in> initHists s2` obtain h' where "h = h' @ (hist s1)" and "set h' \<subseteq> pendingReqs s1" by auto
        moreover from lcp_prop obtain rs where "l_c_p (initHists s2) = rs @ (hist s1)" by auto
        moreover note `x \<notin> set (l_c_p (initHists s2))` and `x \<in> set h`
        ultimately have "x \<in> set h'" by auto
        with `set h' \<subseteq> pendingReqs s1` have "x \<in> pendingReqs s1" by auto
      } moreover
      { fix x c h
        assume "phase s2 c \<noteq> Sleep" and "\<not> initialized s2"
        with P12 have  lcp_prop:"\<exists> rs . l_c_p (initHists s2) = rs @ (hist s1)" 
          by (auto simp add: postfix_all_def linearizations_def)
        assume "x \<notin> set (l_c_p (initHists s2))" and "x \<in> pendingReqs s2"
        from `x \<notin> set (l_c_p (initHists s2))` and lcp_prop have "x \<notin> set (hist s1)" by auto
        moreover have "\<exists> c' . phase s1 c' = Aborted \<and> x = pending s1 c'"
        proof -
          from `x \<in> pendingReqs s2` and `P6 (s1, s2)` 
            obtain c' where "phase s1 c' = Aborted" and "x = pending s2 c'" 
            by (force simp add:pendingReqs_def)
          moreover with `\<not> initialized s2` and  `P7 (s1, s2)` have "x = pending s1 c'" by auto
          ultimately show ?thesis by auto
        qed
        ultimately have "x \<in> pendingReqs s1" by (auto simp add:pendingReqs_def)
      }
      ultimately show ?thesis by (auto simp add:initValidReqs_def)
    qed
  } thus ?thesis by (metis Automata.invariant_def P12_invariant P8_invariant P6_invariant P7_invariant PairE assms)
qed

lemma P1a_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P1a"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P1a (s1, s2)" by simp
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp:"P1a (s1, s2)" and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from reach and P5_invariant have P5:"P5 (s1, s2)" using assms by (metis (full_types) invariantE)
  from P5 and hyp and trans have "P1a (s1', s2')" using assms by auto } 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P1b_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P1b"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P1b (s1, s2)" by simp
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp:"P1b (s1, s2)" and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
  from reach and P1a_invariant have P1a:"P1a (s1, s2)" using assms by (metis (full_types) invariantE)
  from P1a and hyp and trans have "P1b (s1', s2')" using assms by auto } 
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P13_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P13"
proof -
  { fix s1 s2
    assume P11:"P11 (s1, s2)" and P12:"P12 (s1, s2)"
    have "P13 (s1, s2)"
    proof -
      { assume "(\<exists> c . phase s2 c \<noteq> Sleep) \<and> \<not> initialized s2"
        with P12 and P11 obtain rs where 
          initValidReqs_prop:"initValidReqs s2 \<subseteq> pendingReqs s1" 
          and "l_c_p (initHists s2) = rs @ (hist s1)" 
          and "set rs \<subseteq> pendingReqs s1" 
          and "distinct rs" 
          by (auto simp add: postfix_all_def linearizations_def)
        moreover from `l_c_p (initHists s2) = rs @ (hist s1)` 
          have "initValidReqs s2 \<inter> set rs = {}" 
          by (auto simp add:initValidReqs_def)
        ultimately have "postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2)) \<subseteq> postfix_all (hist s1) (linearizations (pendingReqs s1))" 
          by (force simp add: postfix_all_def linearizations_def)
      }
      thus ?thesis by auto
    qed
  }
  thus ?thesis by (metis Automata.invariant_def P12_invariant P11_invariant PairE assms)
qed

lemma P14_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P14"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P14 (s1, s2)" by simp 
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp: "P14 (s1, s2)" and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')" 
    and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  from trans and assms and trans_cases have cases:"trans_cases_p id1 id2 s1 s2 s1' s2'" by blast
  from reach and P6_invariant and P13_invariant and P10_invariant and P2_invariant and P4_invariant 
    have P6:"P6 (s1, s2)" and P13:"P13 (s1, s2)" and P10:"P10 (s1, s2)" and P2:"P2 (s1, s2)" and P4:"P4 (s1, s2)" 
      using assms by (metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE)
  { fix c r
    assume "both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s1 s2 s1' s2'"
    with hyp have "P14 (s1', s2')"  by force
  } moreover
  { fix  c i h r
    assume "i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans2 id1 id2 c id1 h r s1 s1' \<and> Switch_trans1 id1 id2 c id1 h r s2 s2'"
    with hyp have "P14 (s1', s2')" by simp
  } moreover
  { fix  c i h r
    assume "i = id2 \<and> s1 = s1' \<and> Switch_trans2 id1 id2 c id2 h r s2 s2'"
    with hyp have "P14 (s1', s2')" by fastforce
  } moreover
  { fix c i h
    assume "i < id1 \<and> Commit_trans 0 id1 c i h s1 s1'\<and> s2 = s2'"
    with hyp  have "P14 (s1', s2')" by force
  } moreover
  { fix c i h
    assume "i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1'"
    with hyp  have "P14 (s1', s2')" by fastforce
  } moreover
  { fix h 
    assume "Linearize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp and P6 and P2 and P10 have "P14 (s1', s2')" by fastforce
  } moreover
  { fix h 
    assume "Linearize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    with hyp have "P14 (s1', s2')" by (fastforce simp add: linearizations_def postfix_all_def pendingReqs_def)
  } moreover
  { fix h 
    assume a:"Initialize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    have f:"initialized s1" using P10 by auto
    from a and f hyp  have "P14 (s1', s2')" by simp
  } moreover
  { fix h 
    assume "Initialize_trans id1 id2 id1 h s2 s2' \<and> s1 = s1'"
    find_theorems name:P13
    with hyp and P13 have "P14 (s1', s2')" by (auto simp add:linearizations_def pendingReqs_def postfix_all_def) (blast, force)
  } moreover
  { assume "Abort_trans 0 id1 0 s1 s1' \<and> s2 = s2'"
    with hyp have "P14 (s1', s2')" by force
  } moreover
  { fix i
    assume "i = id1 \<and> Abort_trans id1 id2 id1 s2 s2' \<and> s1 = s1'"
    with hyp have "P14 (s1', s2')" by force
  }
  ultimately have "P14 (s1', s2')" using cases by (simp only:trans_cases_p.simps, metis)
}
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

lemma P15_invariant: 
  assumes "id1 < id2" and "id1 \<noteq> 0"
  shows "invariant (composeALMs id1 id2) P15"
proof -
{ fix s1 s2
  assume "(s1, s2) : starts_of (composeALMs id1 id2)" and "0 < id1"
  hence "P15 (s1, s2)" by simp 
} moreover
{ fix s1 s2 s1' s2' act
  assume hyp: "P15 (s1, s2)" and trans:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')" 
    and reach:"reachable (composeALMs id1 id2) (s1, s2)"
  from trans and assms and trans_cases have cases:"trans_cases_p id1 id2 s1 s2 s1' s2'" by blast
  from reach and P1b_invariant and P13_invariant and P10_invariant and P1a_invariant and P5_invariant and P6_invariant
    have P13:"P13 (s1, s2)" and P1b:"P1b (s1, s2)" and P6:"P6 (s1, s2)" and P1a:"P1a (s1, s2)" and P5:"P5 (s1, s2)" and P10:"P10 (s1, s2)"
      using assms by (metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE, metis (full_types) invariantE)
  { fix c r
    assume a:"both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s1 s2 s1' s2'"
    have "P15 (s1', s2')"   
    proof -
      { assume "s1' = s1"
        with hyp and a and assms have ?thesis by auto
      } moreover
      { assume "s1' \<noteq> s1"
        with a and P6 have "s2' = s2" by (simp, metis phase.simps(12) phase.simps(4))
        with `s1' \<noteq> s1` and hyp and a have ?thesis by (force simp add: pendingReqs_def)
      }
      ultimately show ?thesis by auto
    qed
  } moreover
  { fix  c i h r
    assume "i = id1 \<and> id1 \<noteq> 0 \<and> Switch_trans2 id1 id2 c id1 h r s1 s1' \<and> Switch_trans1 id1 id2 c id1 h r s2 s2'"
    with hyp have "P15 (s1', s2')" by (auto simp: pendingReqs_def)
  } moreover
  { fix  c i h r
    assume "i = id2 \<and> s1 = s1' \<and> Switch_trans2 id1 id2 c id2 h r s2 s2'"
    with hyp have "P15 (s1', s2')" by fastforce
  } moreover
  { fix c i h
    assume "i < id1 \<and> Commit_trans 0 id1 c i h s1 s1'\<and> s2 = s2'"
    with hyp  have "P15 (s1', s2')" by (auto simp: pendingReqs_def)
  } moreover
  { fix c i h
    assume "i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1'"
    with hyp  have "P15 (s1', s2')" by fastforce
  } moreover
  { fix h 
    assume "Linearize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    with hyp have "P15 (s1', s2')"  by (auto simp: pendingReqs_def postfix_all_def)
  } moreover
  { fix h 
    assume a:"Linearize_trans id1 id2 id1 h s2 s2'" and "s1 = s1'"
    have "P15 (s1', s2')"
    proof -
      { fix r
        assume "phase s2' (request_snd r) = Sleep" and "r \<in> set (hist s2')" and "r \<notin> pendingReqs s1"
        have "r \<in> set (hist s1)"
        proof -
          from `phase s2' (request_snd r) = Sleep` and a have "phase s2 (request_snd r) = Sleep" by force
          with P1b have "r \<notin> pendingReqs s2" by (auto simp add:pendingReqs_def)
          with a and `r \<in> set (hist s2')` have "r \<in> set (hist s2)" by (auto simp add:postfix_all_def linearizations_def)
          with `phase s2 (request_snd r) = Sleep` and `P15 (s1, s2)` and `r \<notin> pendingReqs s1` show ?thesis by force
        qed
      }
      with `s1 = s1'` show ?thesis by (auto simp: pendingReqs_def)
    qed
  } moreover
  { fix h 
    assume a:"Initialize_trans 0 id1 0 h s1 s1' \<and> s2 = s2'"
    have f:"initialized s1" using P10 by auto
    from a and f hyp  have "P15 (s1', s2')" by simp
  } moreover
  { fix h 
    assume in_init:"Initialize_trans id1 id2 id1 h s2 s2'" and "s1 = s1'"
    have "P15 (s1', s2')"
    proof -
      {
        fix r
        assume "phase s2' (request_snd r) = Sleep" and "r \<in> set (hist s2')" and "r \<notin> pendingReqs s1"
        have "r \<in> set (hist s1)"
        proof -
          from in_init and P13
          have "hist s2' \<in> postfix_all (hist s1) (linearizations (pendingReqs s1))" by fastforce
          with `r \<in> set (hist s2')` have "r \<in> set (hist s1) \<or> r \<in> pendingReqs s1" by (auto simp add: postfix_all_def linearizations_def)
          with `r \<notin> pendingReqs s1` show ?thesis by auto
        qed
      }
      with `s1 = s1'` show ?thesis by (auto simp: pendingReqs_def)
    qed
  } moreover
  { assume "Abort_trans 0 id1 0 s1 s1' \<and> s2 = s2'"
    with hyp have "P15 (s1', s2')" by (force simp: pendingReqs_def)
  } moreover
  { fix i
    assume "i = id1 \<and> Abort_trans id1 id2 id1 s2 s2' \<and> s1 = s1'"
    with hyp have "P15 (s1', s2')" by force
  }
  ultimately have "P15 (s1', s2')" using cases by (simp only:trans_cases_p.simps, metis)
}
ultimately show ?thesis by (metis PairE assms(2) invariantI neq0_conv)
qed

subsection {*The refinement proof*}

definition ref_mapping :: "(ALM_state * ALM_state) => ALM_state" 
  -- {*The refinement mapping between the composition of two ALMs and a single ALM*}
  where
  "ref_mapping \<equiv> \<lambda> (s1, s2) .
     \<lparr>pending = \<lambda>c. (if phase s1 c \<noteq> Aborted then pending s1 c else pending s2 c),
      initHists = {},
      phase = \<lambda>c. (if phase s1 c \<noteq> Aborted then phase s1 c else phase s2 c),
      hist = (if hist s2 = [] then hist s1 else hist s2),
      aborted = aborted s2,
      initialized = True\<rparr>"
declare
  hide_asig_def[simp]
  asig_comp_def[simp]
  ref_mapping_def[simp]
  externals_def[simp]
  
  (*fun_upd_def[simp]*)
  
theorem composition: 
  assumes "id1 \<noteq> 0" and "id1 < id2"
  shows "((composeALMs id1 id2) =<| (ALM_ioa 0 id2))"
  -- {*The composition theorem*}
  proof -
    have same_input_sig:"inp (composeALMs id1 id2) = inp (ALM_ioa 0 id2)" 
      -- {*First we show that both automata have the same input and output signature*}
        using  `id1 \<noteq> 0` and `id1 < id2` by auto
    moreover
    have same_output_sig:"out (composeALMs id1 id2) = out (ALM_ioa 0 id2)" 
      -- {*Then we show that output signatures match*}
      using `id1 \<noteq> 0` and `id1 < id2` 
      by auto
    moreover
    have "traces (composeALMs id1 id2) <= traces (ALM_ioa 0 id2)"
      -- {*Finally we show trace inclusion*}
    find_theorems name:trace_inclusion
    proof - 
      -- {*We use the mapping @{text ref_mapping}, defined before*}
      have "ext (composeALMs id1 id2) = ext (ALM_ioa 0 id2)"  
        -- {*First we show that they have the same external signature*}
        using same_input_sig and same_output_sig by simp
      moreover
      have "is_ref_map ref_mapping (composeALMs id1 id2) (ALM_ioa 0 id2)"
        -- {*Then we show that @{text ref_mapping_comp} is a refinement mapping*}
      find_theorems name:is_ref_map_def 
      proof -
        { -- {*First we show that start states correspond*}
          fix s1 s2
          assume "(s1, s2) : starts_of (composeALMs id1 id2)"
          hence "ref_mapping (s1, s2) : starts_of (ALM_ioa 0 id2)" using  assms by auto
        } moreover 
        {  -- {*Then we show the main property of a refinement mapping*}
          fix s1 s2 s1' s2' act
          assume reachable:"reachable (composeALMs id1 id2) (s1, s2)" 
            and in_trans_comp:"(s1, s2) -act--composeALMs id1 id2-> (s1', s2')"
          txt  {*We make the invariants available for later use*}
          have "P6 (s1, s2)" and "P6 (s1', s2')" and "P9 (s1, s2)" and "P7 (s1, s2)" and "P10 (s1, s2)" and "P4 (s1, s2)" and "P5 (s1, s2)" and "P13 (s1, s2)" and "P1a (s1, s2)" and "P14 (s1, s2)" and "P14 (s1', s2')" and "P15 (s1, s2)" and "P2 (s1, s2)" and "P3 (s1, s2)"
          proof  -
            from reachable and in_trans_comp have "reachable (composeALMs id1 id2) (s1', s2')" using reachable.reachable_n by fast
            with P6_invariant and P9_invariant and P2_invariant and P7_invariant and P10_invariant and P4_invariant and P5_invariant and P13_invariant  and P1a_invariant and P14_invariant and P15_invariant  and P3_invariant `id1 \<noteq> 0` and `id1 < id2` and reachable
            show "P6 (s1, s2)" and "P6 (s1', s2')" and "P9 (s1, s2)" and "P7 (s1, s2)" and "P10 (s1, s2)" and "P4 (s1, s2)" and "P5 (s1, s2)" and "P13 (s1, s2)" and "P1a (s1, s2)" and "P14 (s1, s2)"  and "P14 (s1', s2')"  and "P15 (s1, s2)"  and "P2 (s1, s2)" and "P3 (s1, s2)"
            by (metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def,metis invariant_def)
          qed
          let ?t = "ref_mapping (s1, s2)"
          let ?t' = "ref_mapping (s1', s2')"
          have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'"
            -- {* the main part of the proof *}
          proof -
            { fix c r
              assume act_invoke:"act = Invoke c r"
              have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'"
              proof  -
                let ?ex = "[(Invoke c r, ?t')!]"
                  (*First a bunch of trivial things*)
                have "Finite ?ex" by auto
                moreover have "laststate (?t, ?ex) = ?t'" by (simp add: laststate_def)
                moreover have "mk_trace (ALM_ioa 0 id2)$(?ex) = [Invoke c r!]" by (simp add: mk_trace_def)
                  (*Proof that ?ex relates the mappings*)
                moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                proof -
                  from act_invoke and in_trans_comp and trans_cases_act 
                    have in_invoke:"both id1 id2 (\<lambda> m n . Invoke_trans m n c r) s1 s2 s1' s2'"
                    by simp
                  { assume "s1' \<noteq> s1 & s2' \<noteq> s2" 
                      -- {*contradiction*}
                    with in_invoke and `id1 \<noteq> 0` and `id1 < id2` and `P6 (s1', s2')` have ?thesis by force
                  }
                  moreover
                  { assume a1:"s1' = s1" and a2:"s2' = s2"
                    (*Automatically but slowly:
                    from in_invoke and `P6 (s1,s2)` and `P14 (s1,s2)` and assms and ALM_state_ext have ?thesis
                    by (subst (asm) `s1'=s1`, subst `s1'=s1`, subst (asm) `s2'=s2`, subst `s2'=s2`) (force simp add:ALM_state_ext fun_eq_iff)*)
                    from in_invoke 
                      have pre_s1:"~(phase s1 c = Ready & request_snd r = c & r \<notin> set (hist s1))"
                    by (subst (asm) `s1'=s1`) (force simp: ALM_state_ext fun_eq_iff)
                    moreover from in_invoke 
                      have pre_s2:"~(phase s2 c = Ready & request_snd r = c & r \<notin> set (hist s2))"
                    by (subst (asm) `s2'=s2`) (force simp: ALM_state_ext fun_eq_iff)
                    ultimately have "~(phase ?t c = Ready & request_snd r = c & r \<notin> set (hist ?t))" 
                      using `P14 (s1, s2)` by force
                    hence ?thesis using `id1 \<noteq> 0`  and `s1' = s1` and `s2' = s2` 
                    by fastforce  
                  } moreover
                  { assume "s1' \<noteq> s1" and "s2' = s2"
                    with in_invoke
                      have pre_s1:"phase s1 c = Ready & request_snd r = c & r \<notin> set (hist s1)" 
                      and trans_s1: "s1' = s1\<lparr>pending :=  (pending s1)(c := r), phase := (phase s1)(c := Pending)\<rparr>"
                      by auto
                    have pre_t: "phase ?t c = Ready & request_snd r = c & r \<notin> set (hist ?t)"
                    proof -
                      from pre_s1 have "phase ?t c = Ready & request_snd r = c" by auto
                      moreover have "r \<notin> set (hist ?t)"
                      proof (cases "hist s2 = []")
                        assume "hist s2 = []"
                        with pre_s1 show ?thesis by auto
                      next
                        assume "hist s2 \<noteq> []"
                        show "r \<notin> set (hist ?t)"
                        proof -
                          { assume "r \<in> set (hist ?t)"
                            with `hist s2 \<noteq> []` have "r \<in> set (hist s2)" by auto
                            moreover from pre_s1 and `P6 (s1, s2)` have "phase s2 (request_snd r) = Sleep" by force
                            moreover note `P15 (s1, s2)`
                            ultimately have "r \<in> set (hist s1) \<or> r \<in> pendingReqs s1" by auto 
                            with pre_s1 have "r \<in> pendingReqs s1" by auto
                            with `P1a (s1, s2)` and pre_s1 have False by (auto simp add:pendingReqs_def)
                          }
                          thus ?thesis by auto
                        qed
                      qed
                      ultimately show ?thesis by auto
                    qed
                    moreover from pre_s1 and trans_s1 and `s2' = s2` 
                      have trans_t:"?t' = ?t\<lparr>pending :=  (pending ?t)(c := r), phase := (phase ?t)(c := Pending)\<rparr>" by fastforce
                    ultimately have ?thesis by auto
                  }
                  moreover
                  { assume "s1' = s1" and "s2' \<noteq> s2"
                    (*Automatically but slowly:
                    from in_invoke and `P6 (s1,s2)` and `P3 (s1,s2)` and `P14 (s1,s2)` and assms and ALM_state_ext have ?thesis
                    by (subst (asm) `s1'=s1`, subst `s1'=s1`) (force simp add:ALM_state_ext fun_eq_iff)*)
                    with in_invoke and `id1 \<noteq> 0` 
                      have pre_s2: "phase s2 c = Ready & request_snd r = c & r \<notin> set (hist s2)" and 
                      trans_s2: "s2' = s2\<lparr>pending :=  (pending s2)(c := r), phase := (phase s2)(c := Pending)\<rparr>" 
                      by auto
                    from pre_s2 and `P6 (s1, s2)`  have aborted_s1_c:"phase s1 c = Aborted" by force 
                    with pre_s2 and `P3 (s1, s2)` and `P14 (s1, s2)` 
                      have pre_t:"phase ?t c = Ready & request_snd r = c & r \<notin> set (hist ?t)" by auto
                    moreover have trans_t:"?t' = ?t\<lparr>pending :=  (pending ?t)(c := r), phase := (phase ?t)(c := Pending)\<rparr>" 
                      using aborted_s1_c and `s1' = s1` and trans_s2 by auto
                    ultimately have ?thesis by auto 
                  }
                  ultimately show ?thesis by fast
                qed
                ultimately show ?thesis using act_invoke by (metis is_exec_frag_cons transition_is_ex) 
              qed
            } moreover
            { fix  c i h r
              assume act_switch:"act = Switch c i h r"
              have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'"
              proof -
                from act_switch and in_trans_comp 
                  have cases:"i = id1 \<or> i = id2" by simp
                { assume a:"i = id1"
                  hence b:"act \<notin> ext (ALM_ioa 0 id2)" using act_switch and assms by simp
                  let ?ex = "nil"
                  have "mk_trace (ALM_ioa 0 id2)\<cdot>?ex = nil" 
                    using act_switch and assms by (simp add:mk_trace_def)
                  moreover have "laststate (?t, ?ex) = ?t" by simp
                  moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)" by simp
                  moreover have "?t' = ?t"
                  proof -
                    from act_switch and a and in_trans_comp and trans_cases_act and assms
                      have in_switch:"Switch_trans2 id1 id2 c id1 h r s1 s1' \<and> Switch_trans1 id1 id2 c id1 h r s2 s2'" by force
                    from in_switch and `id1 \<noteq> 0` 
                      have pre_s1:"aborted s1 \<and> phase s1 c = Pending \<and> r = pending s1 c \<and> (if initialized s1 then (h \<in> postfix_all (hist s1) (linearizations (pendingReqs s1))) else (h : postfix_all (l_c_p (initHists s1)) (linearizations (initValidReqs s1))))" 
                      and trans_s1: "s1' = s1\<lparr>phase := (phase s1)(c := Aborted)\<rparr>"
                      by auto
                    have pre_s2:"phase s2 c = Sleep" and trans_s2: "s2' = s2 \<lparr>initHists := {h} \<union> (initHists s2), phase := (phase s2)(c := Pending), pending := (pending s2)(c := r)\<rparr>"
                    proof -
                      from pre_s1 have "phase s1 c = Pending" by auto
                      with `P6 (s1, s2)` have "phase s2 c = Sleep" by fastforce
                      with in_switch and `id1 \<noteq> 0` and `id1 < id2` 
                        show "phase s2 c = Sleep" 
                        and  "s2' = s2 \<lparr>initHists := {h} \<union> (initHists s2), phase := (phase s2)(c := Pending), pending := (pending s2)(c := r)\<rparr>" 
                          by auto
                    qed
                    from pre_s1 and pre_s2 and trans_s1 and trans_s2 and `P1a (s1, s2)` 
                      have "pending ?t c = pending ?t' c & initHists ?t = initHists ?t' & hist ?t = hist ?t' & aborted ?t = aborted ?t' \<and> phase ?t' c = phase ?t c" 
                      by simp
                    moreover note pre_s1 and pre_s2 and trans_s1 and trans_s2
                    ultimately show ?thesis by (force simp add: fun_eq_iff)
                  qed
                  ultimately have ?thesis using b by (metis nothing_is_ex)
                }
                moreover
                { assume a:"i = id2"
                  hence b:"act \<in> ext (ALM_ioa 0 id2)" using act_switch and assms by simp
                  let ?ex = "[(Switch c i h r, ?t')!]"
                  have "Finite ?ex" by auto
                  moreover have "laststate (?t, ?ex) = ?t'" by (simp add: laststate_def)
                  moreover have "mk_trace (ALM_ioa 0 id2)$(?ex) = [Switch c i h r!]" 
                    using act_switch and assms and a by (simp add: mk_trace_def)
                  moreover have exec:"is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                  proof -
                    from act_switch and a and in_trans_comp and trans_cases_act and assms
                      have in_switch:"s1 = s1' \<and> Switch_trans2 id1 id2 c id2 h r s2 s2'" by force
                    from in_switch and `id1 < id2` have "s1' = s1" by auto
                    from `id1 \<noteq> 0` and `id1 < id2` in_switch 
                      have pre_s2:"aborted s2 & phase s2 c = Pending & r = pending s2 c & (if initialized s2 then (h \<in> postfix_all (hist s2) (linearizations (pendingReqs s2))) else (h : postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2))))" 
                      and trans_s2: "s2' = s2\<lparr>phase := (phase s2)(c := Aborted)\<rparr>" 
                      by auto
                    from pre_s2 
                      have s1_aborted:"phase s1 c = Aborted" using `P6 (s1, s2)` by auto
                    have pre_t:"aborted ?t & phase ?t c = Pending & initialized ?t & h : postfix_all (hist ?t) (linearizations (pendingReqs ?t)) & r = pending ?t c" 
                    proof -
                      from s1_aborted and pre_s2 have "aborted ?t & pending ?t c = r" 
                        and "phase ?t c = Pending" and "initialized ?t" by (auto simp add: fun_eq_iff)
                      moreover have "h : postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                      proof -
                        { assume case1_1:"initialized s2" 
                            and case1_2:"h : postfix_all (hist s2) (linearizations (pendingReqs s2))"
                          hence "suffixeq (hist s1) (hist s2)" using `P14 (s1, s2)` 
                            by (auto simp: suffixeq_def)
                          have "h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                          proof -
                            have "hist ?t = hist s2"
                            proof (cases "hist s2 = []")
                              assume "hist s2 = []"
                              show "hist ?t = hist s2"
                              proof -
                                from `hist s2 = []` and `suffixeq (hist s1) (hist s2)` 
                                  have "hist s1 = []" by (auto simp add:suffixeq_def)
                                with `hist s2 = []` show  "hist ?t = hist s2" by auto
                              qed
                            next
                              assume "hist s2 \<noteq> []"
                              thus "hist ?t = hist s2" by simp
                            qed
                            moreover have "pendingReqs s2 <= pendingReqs ?t"
                            proof - 
                              { fix r
                                assume "r \<in> pendingReqs s2"
                                with this obtain c where "r = pending s2 c" 
                                  and "pending s2 c \<notin> set (hist s2)" 
                                  and "phase s2 c \<in> {Pending, Aborted}" 
                                  by (force simp: pendingReqs_def)
                                moreover with `P6 (s1, s2)` have "phase s1 c = Aborted" by auto
                                moreover note `suffixeq (hist s1) (hist s2)`
                                ultimately have "\<exists>ca. r = pending ?t ca \<and> r \<notin> set (hist ?t) \<and> (phase ?t ca = Pending \<or> phase ?t ca = Aborted)"
                                  by force
                                hence "r \<in> pendingReqs ?t" by (simp add: pendingReqs_def)
                              }
                              thus ?thesis by auto
                            qed
                            moreover note case1_2
                            ultimately show ?thesis by (auto simp add: linearizations_def postfix_all_def)
                          qed
                        } moreover 
                        { assume case2_1:"\<not> initialized s2" 
                          and case2_2:"h : postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2))"
                          from case2_1 and `P10 (s1, s2)` have "hist s2 = []" by force
                          have "h : postfix_all (hist s1) (linearizations (pendingReqs s1))"
                          proof -
                            from pre_s2 have "phase s2 c \<noteq> Sleep" by auto
                            moreover note `P13 (s1, s2)` and case2_1 and case2_2
                            ultimately  show ?thesis by auto
                          qed
                          moreover from `hist s2 = []` have "hist ?t = hist s1" by auto
                          moreover have "pendingReqs ?t = pendingReqs s1"
                          proof -
                            { fix r
                              assume "r \<in> pendingReqs ?t"
                              with this obtain c' where "r = pending ?t c'" and "r \<notin> set (hist ?t)" 
                              and "phase ?t c' \<in> {Pending, Aborted}"
                                by (force simp add:pendingReqs_def)
                              have "r \<in> pendingReqs s1"
                              proof (cases "phase s1 c' = Aborted")
                                assume "phase s1 c' = Aborted"
                                with `phase ?t c' \<in> {Pending, Aborted}` and `r = pending ?t c'` 
                                have "phase s2 c' \<in> {Pending, Aborted}" and "r = pending s2 c'" by auto
                                with `P6 (s1, s2)` and case2_1 and `P7 (s1, s2)` 
                                and `hist ?t = hist s1` and `r \<notin> set (hist ?t)` 
                                  have "phase s1 c' = Aborted" and "r = pending s1 c'" and "r \<notin> set (hist s1)" 
                                  by (auto, force, force, force, force)
                                thus ?thesis by (auto simp add:pendingReqs_def)
                              next
                                assume "phase s1 c' \<noteq> Aborted"
                                with `r = pending ?t c'` and `r \<notin> set (hist ?t)` 
                                and `phase ?t c' \<in> {Pending, Aborted}` and `hist ?t = hist s1` 
                                show ?thesis by (auto simp add:pendingReqs_def)
                              qed
                            } moreover
                            { fix r
                              assume "r \<in> pendingReqs s1"
                              with this obtain c where "r = pending s1 c" 
                              and "phase s1 c \<in> {Pending, Aborted}" and "r \<notin> set (hist s1)" 
                              by (auto simp add:pendingReqs_def)
                              with `hist s2 = []` and `\<not> initialized s2` and `P7 (s1, s2)` 
                              have "r \<in> pendingReqs ?t" by (auto simp add:pendingReqs_def)
                            }
                            ultimately show ?thesis by blast
                          qed
                          ultimately have ?thesis by (auto simp add: postfix_all_def linearizations_def)
                        }
                        moreover from pre_s2 have "(if initialized s2 then (h : postfix_all (hist s2) (linearizations (pendingReqs s2))) else (h : postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2))))" by auto
                        ultimately show ?thesis by auto   
                      qed
                      ultimately show ?thesis by auto
                    qed
                    moreover have trans_t:"?t' = ?t\<lparr>phase := (phase ?t)(c := Aborted)\<rparr>" 
                      using s1_aborted and `s1' = s1` and trans_s2 by auto 
                    ultimately have "Switch_trans2 0 id2 c id2 h r ?t ?t'" by force
                    thus ?thesis using assms and act_switch and a by force
                  qed
                  ultimately have ?thesis using b and act_switch by (metis is_exec_frag_cons transition_is_ex)
                }
                ultimately show ?thesis using cases by auto
              qed
            } moreover 
            { fix c i h
              assume commit_act:"act = Commit c i h"
              have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'"
              proof -
                let ?ex = "[(Commit c i h, ?t')!]"
                from commit_act and in_trans_comp
                  have cases:"(i < id1 \<and> Commit_trans 0 id1 c i h s1 s1'\<and> s2 = s2') \<or> (i \<ge> id1 \<and> i < id2 \<and> Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1')"
                  by auto
                moreover have "Finite ?ex" by auto
                moreover have "laststate (?t, ?ex) = ?t'" by (simp add: laststate_def)
                moreover have "mk_trace (ALM_ioa 0 id2)$(?ex) = [Commit c i h!]" 
                  using cases and commit_act and assms by (auto simp add:mk_trace_def)
                moreover
                { assume "i < id1" and in_commit:"Commit_trans 0 id1 c i h s1 s1' \<and> s2 = s2'"
                  from `i < id1` and commit_act and assms 
                    have "act \<in> ext (ALM_ioa 0 id2)" by simp
                  moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                  proof -
                    from `i < id1` and in_commit have "s2' = s2" 
                      and pre_s1:"phase s1 c = Pending \<and> pending s1 c \<in> set (hist s1) \<and> h = dropWhile (\<lambda> r . r \<noteq> pending s1 c) (hist s1)" 
                      and trans_s1:"s1' = s1 \<lparr>phase := (phase s1)(c := Ready)\<rparr>"
                      by auto
                    from pre_s1 have s1_not_aborted_c:"phase s1 c \<noteq> Aborted" by auto
                    have pre_t:"phase ?t c = Pending & pending ?t c \<in> set (hist ?t) \<and> h = dropWhile (\<lambda> r . r \<noteq> pending ?t c) (hist ?t)"
                    proof (cases "hist s2 = []")
                      assume "hist s2 = []"
                      with pre_s1 and `phase s1 c \<noteq> Aborted` show ?thesis by auto
                    next
                      assume "hist s2 \<noteq> []"
                      hence "initialized s2" using `P10 (s1, s2)` by auto
                      from pre_s1 and `phase s1 c \<noteq> Aborted` have "phase ?t c = Pending" 
                        and "pending ?t c = pending s1 c" and "pending s1 c \<in> set (hist s1)" 
                        by auto
                      moreover have "pending ?t c \<in> set (hist ?t)"
                      proof -
                        from `initialized s2` and `P14 (s1, s2)` obtain rs3 where "hist s2 = rs3 @ (hist s1)" by auto
                        with `pending s1 c \<in> set (hist s1)` and `hist s2 = rs3 @ (hist s1)` 
                          and `pending ?t c = pending s1 c` 
                          show "pending ?t c \<in> set (hist ?t)" by auto
                      qed
                      moreover have "h = dropWhile (\<lambda> r . r \<noteq> pending ?t c) (hist ?t)"
                      proof -
                        from `pending s1 c \<in> set (hist s1)` obtain rs1 rs2 where 
                          "hist s1 = rs2 @ rs1" and "hd rs1 = pending s1 c" 
                          and "rs1 \<noteq> []" and "pending s1 c \<notin> set rs2"  
                          by (metis hd.simps in_set_conv_decomp_first list.simps(3))
                        with `pending ?t c = pending s1 c` 
                          and dropWhile_lemma[of "hist s1" rs1 "pending s1 c"] 
                          and pre_s1 have "h = rs1" by auto
                        moreover have "dropWhile (\<lambda> r . r \<noteq> pending ?t c) (hist ?t) = rs1"
                        proof -
                          from `initialized s2` and `P14 (s1, s2)` obtain rs3 where 
                            "hist s2 = rs3 @ (hist s1)" and "set rs3 \<inter> set (hist s1) = {}" by auto
                          with `pending s1 c \<in> set (hist s1)` and `hist s1 = rs2 @ rs1` 
                            have "hist s2 = rs3 @ rs2 @ rs1" and "pending s1 c \<notin> set rs3" by auto
                          with `pending s1 c \<notin> set rs2` obtain rs4 where 
                            "hist s2 = rs4 @ rs1" and "pending s1 c \<notin> set rs4" by auto
                          with `hd rs1 = pending s1 c` and `rs1 \<noteq> []` 
                            and dropWhile_lemma[of "hist s2" rs1 "pending s1 c"] 
                            have "dropWhile (\<lambda> r . r \<noteq> pending s1 c) (hist s2) = rs1" by auto
                          thus ?thesis using `hist s2 \<noteq> []` and `pending ?t c = pending s1 c` by auto
                        qed
                        ultimately show ?thesis by auto
                      qed
                      ultimately show ?thesis by auto
                    qed
                    moreover from `s2' = s2` and s1_not_aborted_c and trans_s1 
                      have trans_t:"?t' = ?t \<lparr>phase := (phase ?t)(c := Ready)\<rparr>" by (simp add:fun_eq_iff)
                    ultimately show ?thesis using `id1 < id2` by auto
                  qed
                } moreover
                { assume "i \<ge> id1 \<and> i < id2" and in_commit:"Commit_trans id1 id2 c i h s2 s2' \<and> s1 = s1'"
                  from `i \<ge> id1 \<and> i < id2` and commit_act and assms 
                    have "act \<in> ext (ALM_ioa 0 id2)" by simp
                  moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                  proof -
                    from `i \<ge> id1 \<and> i < id2` and in_commit have "s1' = s1" 
                      and pre_s2:"phase s2 c = Pending \<and> pending s2 c \<in> set (hist s2) \<and> h = dropWhile (\<lambda> r . r \<noteq> pending s2 c) (hist s2)" 
                      and trans_s2:"s2' = s2 \<lparr>phase := (phase s2)(c := Ready)\<rparr>" 
                      by auto
                    from pre_s2 and `P6 (s1, s2)` 
                      have facts:"aborted s1 & phase s1 c = Aborted & hist s2 \<noteq> []" 
                      by force
                    with pre_s2 have pre_t:"phase ?t c = Pending \<and> pending ?t c \<in> set (hist ?t) \<and> h = dropWhile (\<lambda> r . r \<noteq> pending ?t c) (hist ?t)" by auto
                    moreover from `s1' = s1` and facts 
                      and trans_s2 have trans_t:"?t' = ?t \<lparr>phase := (phase ?t)(c := Ready)\<rparr>" by (auto simp add:fun_eq_iff)
                    ultimately show ?thesis using `id1 < id2` by auto
                  qed
                }
                ultimately show ?thesis using commit_act by (metis is_exec_frag_cons transition_is_ex) 
              qed
            } moreover
            { fix i h
              assume linearize_act:"act = Linearize i h"
              have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'"
              proof -
                let ?ex = "[(Linearize 0 h, ?t')!]"
                from linearize_act and assms 
                  have "act \<notin> ext (ALM_ioa 0 id2)" by simp
                moreover have "Finite ?ex" by auto
                moreover have "laststate (?t, ?ex) = ?t'" by (simp add: laststate_def)
                moreover have "mk_trace (ALM_ioa 0 id2)\<cdot>?ex = nil" by (simp add: mk_trace_def)
                moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                proof -
                  from linearize_act and in_trans_comp
                    have cases:"(Linearize_trans 0 id1 i h s1 s1' \<and> s2 = s2') \<or> (Linearize_trans id1 id2 i h s2 s2' \<and> s1 = s1')"
                    using assms by auto
                  moreover
                  { assume in_lin:"Linearize_trans 0 id1 i h s1 s1'" and "s2 = s2'"
                    from in_lin and assms 
                      have pre_s1:"initialized s1 & ~ aborted s1 & h \<in> postfix_all (hist s1) (linearizations (pendingReqs s1))" 
                      and trans_s1:"s1' = s1\<lparr>hist := h, initialized := True\<rparr>"
                      by auto
                    have pre_t:"initialized ?t & ~ aborted ?t & h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                    proof -
                      from  pre_s1 have "~ aborted s1" by auto
                      with `P9 (s1, s2)` have "~ aborted ?t" and "initialized ?t" by auto
                      moreover have "h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                      proof -
                        from `\<not> aborted s1` have "hist ?t = hist s1" 
                          using `P6 (s1, s2)` and `P2 (s1, s2)`  by auto
                        moreover have "pendingReqs s1 \<subseteq> pendingReqs ?t" 
                        proof -
                          { fix x
                            assume "x \<in> pendingReqs s1"
                            moreover note `\<not> aborted s1` and `P6 (s1 ,s2)`
                            ultimately obtain c where "x = pending s1 c" 
                              and "phase s1 c = Pending" and "pending s1 c \<notin> set (hist s1)" 
                              by (auto simp add:pendingReqs_def)
                            hence "x \<in> pendingReqs ?t" using `hist ?t = hist s1` 
                              by (force simp add: pendingReqs_def)
                          }
                          thus ?thesis by auto
                        qed
                        moreover from pre_s1 
                          have "h \<in> postfix_all (hist s1) (linearizations (pendingReqs s1))" by auto
                        ultimately show ?thesis by (auto simp add: postfix_all_def linearizations_def)
                      qed
                      ultimately show ?thesis by auto
                    qed
                    moreover have trans_t: "?t' = ?t\<lparr>hist := h, initialized := True\<rparr>"
                    proof -
                      have "hist ?t' = hist s1'"
                      proof -
                        from pre_s1 have "~ aborted s1" by auto
                        with  `P6 (s1, s2)` and `P2 (s1, s2)` have "hist s2 = []" by auto
                        with `s2 = s2'` show ?thesis by auto
                      qed
                      with trans_s1 have "hist ?t' = h" by auto
                      thus ?thesis using `s2 = s2'` and trans_s1 by (auto simp add:fun_eq_iff)
                    qed
                    ultimately have ?thesis  by auto
                  } moreover
                  { assume in_lin:"Linearize_trans id1 id2 i h s2 s2'" and "s1 = s1'"
                    from in_lin and assms  
                      have pre_s2: "initialized s2 \<and> \<not> aborted s2 \<and> h \<in> postfix_all (hist s2) (linearizations (pendingReqs s2))" and trans_s2: "s2' = s2\<lparr>hist := h\<rparr>" 
                      by auto
                    have pre_t:"initialized ?t \<and> \<not> aborted ?t \<and> h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                    proof -
                      have  "\<not> aborted ?t" and "initialized ?t" using pre_s2 by auto
                      moreover have "h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                      proof -
                        from pre_s2 have "initialized s2" by auto
                        hence "suffixeq (hist s1) (hist s2)" using `P14 (s1, s2)` 
                          by (auto simp add: suffixeq_def)
                        hence "hist ?t = hist s2"  by auto
                        moreover have "pendingReqs s2 \<subseteq> pendingReqs ?t"
                        proof -
                          { fix x
                            assume "x \<in> pendingReqs s2"
                            from this obtain c where "x = pending s2 c" 
                              and "phase s2 c \<in> {Pending, Aborted}" and "pending s2 c \<notin> set (hist s2)" 
                              by (auto simp add:pendingReqs_def)
                            with `P6 (s1, s2)` and `hist ?t = hist s2` 
                            have "x \<in> pendingReqs ?t" by (force simp add: pendingReqs_def)
                          } thus ?thesis by auto
                        qed
                        moreover from pre_s2 
                          have "h \<in> postfix_all (hist s2) (linearizations (pendingReqs s2))" by auto
                        ultimately show ?thesis by (auto simp add:postfix_all_def linearizations_def)
                      qed
                      ultimately show ?thesis by auto
                    qed
                    moreover have trans_t: "?t' = ?t\<lparr>hist := h\<rparr>"
                    proof -
                      from pre_s2 and trans_s2 have "initialized s2'" by auto
                      hence "suffixeq (hist s1') (hist s2')"  
                        using `P14 (s1', s2')` by (auto simp add:suffixeq_def)
                      hence "hist ?t' = hist s2'" by auto
                      with trans_s2 and `s1 = s1'` show ?thesis by (auto simp add: fun_eq_iff)
                    qed
                    ultimately have ?thesis by fastforce  
                  } ultimately show ?thesis by auto
                qed
                ultimately show ?thesis using linearize_act by (metis (hide_lams, no_types) move_def)
              qed
            } moreover
            { fix i h 
              assume init_act:"act = Initialize i h"
              have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'"
              proof -
                let ?ex = "[(Linearize 0 h, ?t')!]"
                from init_act and assms 
                  have "act \<notin> ext (ALM_ioa 0 id2)" by simp
                moreover have "Finite ?ex" by auto
                moreover have "laststate (?t, ?ex) = ?t'" by (simp add: laststate_def)
                moreover have "mk_trace (ALM_ioa 0 id2)\<cdot>?ex = nil" by (simp add: mk_trace_def)
                moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                proof -
                  from init_act and in_trans_comp
                    have cases:"(Initialize_trans 0 id1 i h s1 s1' \<and> s2 = s2') \<or> (Initialize_trans id1 id2 i h s2 s2' \<and> s1 = s1')"
                    using assms by auto
                  moreover
                  { assume in_initialize:"Initialize_trans 0 id1 i h s1 s1'" and "s2' = s2"
                    hence False  using `P10 (s1, s2)` by auto 
                    hence ?thesis by  auto
                  } moreover
                  { assume in_initialize:"Initialize_trans id1 id2 i h s2 s2'" and "s1' = s1"
                    from in_initialize and assms 
                      have pre_s2:"(\<exists> c . phase s2 c \<noteq> Sleep) \<and> \<not> aborted s2 \<and> \<not> initialized s2 \<and> h \<in> postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2))" 
                      and trans_s2:"s2' = s2\<lparr>hist := h, initialized := True\<rparr>"
                      by (auto, metis awake.simps(1))
                    have pre_t:"initialized ?t \<and> \<not> aborted ?t \<and> h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                    proof -
                      from pre_s2 have "initialized ?t \<and> \<not> aborted ?t" by auto
                      moreover have "h \<in> postfix_all (hist ?t) (linearizations (pendingReqs ?t))"
                      proof -
                        from pre_s2 have "h \<in> postfix_all (l_c_p (initHists s2)) (linearizations (initValidReqs s2))" 
                          and "\<not> initialized s2" and "\<exists> c . phase s2 c \<noteq> Sleep" by auto
                        with `P13 (s1, s2)` 
                          have "h \<in> postfix_all (hist s1) (linearizations (pendingReqs s1))" by auto
                        moreover from `\<not> initialized s2` and `P10 (s1, s2)` 
                          have "hist ?t = hist s1" by auto
                        moreover have "pendingReqs s1 \<subseteq> pendingReqs ?t"
                        proof -
                          { fix x
                            assume "x \<in> pendingReqs s1"
                            from this obtain c where "x = pending s1 c" 
                              and "phase s1 c \<in> {Pending, Aborted}" 
                              and "pending s1 c \<notin> set (hist s1)" by (auto simp add:pendingReqs_def)
                            have "x \<in> pendingReqs ?t"
                            proof (cases "phase s1 c = Pending")
                              assume "phase s1 c = Pending"
                              with `x = pending s1 c` and `pending s1 c \<notin> set (hist s1)` 
                                and `hist ?t = hist s1` show ?thesis by (force simp add:pendingReqs_def)
                            next
                              assume "phase s1 c \<noteq> Pending"
                              with `phase s1 c \<in> {Pending, Aborted}` have "phase s1 c = Aborted" by auto
                              with `\<not> initialized s2` and `P6 (s1, s2)` and `P7 (s1, s2)` 
                                have "pending s2 c = pending s1 c" 
                                and "phase s2 c \<in> {Pending, Aborted}" by auto
                              with `x = pending s1 c` and `pending s1 c \<notin> set (hist s1)` 
                              and `hist ?t = hist s1` and `P6 (s1, s2)` 
                              show ?thesis by (auto simp add:pendingReqs_def) 
                            qed
                          } thus ?thesis by auto
                        qed
                        ultimately show ?thesis by (auto simp add:postfix_all_def linearizations_def)
                      qed
                      ultimately show ?thesis by auto
                    qed
                    moreover have trans_t:"?t' = ?t\<lparr>hist := h\<rparr>"
                    proof -
                      from pre_s2 have "\<exists> c . phase s2 c \<noteq> Sleep" by auto
                      with trans_s2 have "initialized s2'" and "\<exists> c . phase s2' c \<noteq> Sleep" by auto
                      hence "suffixeq (hist s1') (hist s2')" using `P14 (s1', s2')` 
                        by (auto simp add:suffixeq_def)
                      hence "hist ?t' = hist s2'" by auto
                      with trans_s2 and `s1' = s1` show ?thesis by (auto simp add:fun_eq_iff)
                    qed
                    ultimately have ?thesis by auto
                  } 
                  ultimately show ?thesis by fast
                qed
                ultimately show ?thesis using init_act by (metis (hide_lams, no_types) move_def)
              qed
            } moreover
            { fix i
              assume abort_act:"act = Abort i"
              from abort_act and assms 
                  have "act \<notin> ext (ALM_ioa 0 id2)" by simp
              from abort_act and in_trans_comp
                have cases:"(i = 0 \<and> Abort_trans 0 id1 i s1 s1' \<and> s2 = s2') \<or> (i = id1 \<and> Abort_trans id1 id2 i s2 s2' \<and> s1 = s1')"
                using assms by auto
              moreover
              { assume in_abort:"Abort_trans 0 id1 i s1 s1'" and "s2 = s2'" and "i = 0"
                let ?ex = "nil"
                have "Finite ?ex" by auto
                moreover have "laststate (?t, ?ex) = ?t" by (simp add: laststate_def)
                moreover have "mk_trace (ALM_ioa 0 id2)\<cdot>?ex = nil" using `id1 < id2` by (simp add: mk_trace_def)
                moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)" by (auto simp add:is_exec_frag_def)
                moreover have "?t' = ?t"
                proof -
                  from assms and in_abort and `P6 (s1, s2)` and `P2 (s1, s2)` 
                    have "\<forall> c . phase s1 c \<noteq> Aborted" and "hist s2 = []" 
                    and "\<forall> c . phase s2 c = Sleep" 
                    by (auto simp add:fun_eq_iff)
                  with assms and in_abort and `s2=s2'` show ?thesis by (auto simp add:fun_eq_iff)
                qed
                ultimately have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'" 
                  using abort_act and assms using `act \<notin> ext (ALM_ioa 0 id2)` by (metis nothing_is_ex)
              } moreover
              { assume in_abort:"Abort_trans id1 id2 id1 s2 s2'" and "s1' = s1" and "i = id1"
                let ?ex = "[(Abort 0, ?t')!]"
                have "Finite ?ex" by auto
                moreover have "laststate (?t, ?ex) = ?t'" by (simp add: laststate_def)
                moreover have "mk_trace (ALM_ioa 0 id2)\<cdot>?ex = nil" by (simp add: mk_trace_def)
                moreover have "is_exec_frag (ALM_ioa 0 id2) (?t, ?ex)"
                proof -
                  from in_abort and `id1 \<noteq> 0`
                    have pre_s2:"~ aborted s2 & (\<exists> c . phase s2 c \<noteq> Sleep)" 
                    and trans_s2:"s2' = s2\<lparr>aborted:= True\<rparr>" 
                    by auto
                  from pre_s2 and `P6 (s1, s2)` 
                    have pre_t:"~ aborted ?t  & (\<exists> c . phase ?t c \<noteq> Sleep)" 
                    by auto
                  moreover from trans_s2 and `s1' = s1` 
                    have trans_t:"?t' = ?t\<lparr>aborted:= True\<rparr>" by (auto simp add: fun_eq_iff)
                  ultimately show ?thesis  by auto
                qed
                ultimately have "EX ex. move (ALM_ioa 0 id2) ex ?t act ?t'" 
                  using assms using `act \<notin> ext (ALM_ioa 0 id2)` by (metis (hide_lams, no_types) move_def) 
              }
              ultimately have ?thesis using abort_act by fast
            }
            ultimately show ?thesis by (cases act, auto)
          qed
        } 
        ultimately show ?thesis by (auto simp only:is_ref_map_def)
      qed
      ultimately show ?thesis by (metis trace_inclusion)
    qed
  ultimately show ?thesis by (metis ioa_implements_def)
qed

end
